#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import argparse
import logging
import pathlib
from typing import Tuple, Union

import sys
from distutils.util import strtobool

from ingest_envoy.interfaces import IngestLauncherIF
from ingest_envoy.launchers import IngestCalibrationLauncher, IngestImageLauncher
from ingest_envoy.solicitor import Solicitor
from pycapo import CapoConfig

from ingest_envoy.utilities import IngestType, VLASSIngestType

"""
Setup and Launch ingestion via Workspaces
"""

logger = logging.getLogger("ingest_envoy")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


def _get_settings(
    arg_type: Union[IngestType, VLASSIngestType], filename: str = None, cal_spl: str = None, source_dir: str = None
) -> dict:
    """
    Determine needed settings from metadata.json file

    :param arg_type: the type of ingestion to run
    :param filename: the provided metadata file
    :param cal_spl: the calibration science product locator to use (needed for VLASS image ingestion)
    :param source_dir: the product source directory to ingest (needed for VLASS ingestion)
    :return: a dict containing required settings to run the provided ingestion type
    """
    ingestion_settings = CapoConfig().settings("edu.nrao.workspaces.IngestionSettings")
    capability_url = CapoConfig().settings("edu.nrao.workspaces.CapabilitySettings").externalServiceUrl
    workflow_url = CapoConfig().settings("edu.nrao.workspaces.WorkflowSettings").serviceUrl

    parameters = Solicitor(arg_type, [capability_url, workflow_url], filename=filename).solicit_parameters()

    if cal_spl is not None:
        parameters["calSpl"] = cal_spl
    if source_dir is not None:
        parameters["workflowDir"] = pathlib.Path(source_dir).name
        parameters["seciCachePath"] = pathlib.Path(source_dir).parent

    parameters["staging_area"] = ingestion_settings.stagingDirectory
    parameters["storage_area"] = ingestion_settings.storageDirectory
    parameters["useIngest"] = strtobool(ingestion_settings.useIngest)

    return parameters


def arg_parser() -> argparse.ArgumentParser:
    """
    Create option parser for this envoy

    :return: argument parser
    """
    parser = argparse.ArgumentParser(
        description="Workspaces Ingestion System", formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument(
        "--calibration",
        nargs=1,
        action="store",
        required=False,
        help="run ingestion for a calibration product",
    )
    parser.add_argument(
        "--image",
        nargs=1,
        action="store",
        required=False,
        help="run ingestion for an image product",
    )
    parser.add_argument(
        "--seci",
        nargs=2,
        action="store",
        required=False,
        help="run ingestion for VLASS SECI image products",
    )
    return parser


def check_ingest_type(args_type: Union[IngestType, VLASSIngestType], parameters: dict) -> bool:
    """
    Ensure that we are running the correct type of ingestion for a given product

    :param args_type: type of ingestion requested
    :param parameters: the parameters determined from metadata.json
    :return: boolean representing if the requested ingestion type matches the product type
    """
    wf_name = parameters["workflowName"]
    if args_type.value in wf_name:
        return True
    elif args_type == IngestType.IMG and "imaging" in wf_name:
        return True
    return False


def _determine_ingestion_info(
    args: argparse.Namespace,
) -> Tuple[Union[IngestType, VLASSIngestType], dict, IngestLauncherIF]:
    """
    Determine parameters and launcher needed for ingestion

    :param args: args dict from parser
    :return: tuple containing the ingestion type, parameters dict, and launcher object
    """
    arg_type = None
    parameters = None
    launcher = None

    if args.calibration is not None:
        arg_type = IngestType.CAL
        parameters = _get_settings(arg_type, filename=args.calibration[0])
        launcher = IngestCalibrationLauncher(parameters)
    elif args.image is not None:
        arg_type = IngestType.IMG
        parameters = _get_settings(arg_type, filename=args.image[0])
        launcher = IngestImageLauncher(arg_type, parameters)
    elif args.seci is not None:
        arg_type = VLASSIngestType.SECI
        parameters = _get_settings(arg_type, cal_spl=args.seci[0], source_dir=args.seci[1])
        launcher = IngestImageLauncher(arg_type, parameters)

    return arg_type, parameters, launcher


def main():
    """
    Entry point for envoy

    :return:
    """
    args = arg_parser().parse_args()

    arg_type, parameters, launcher = _determine_ingestion_info(args)

    if check_ingest_type(args_type=arg_type, parameters=parameters):
        result = launcher.launch_ingestion()
        if result == 0:
            # Ingestion succeeded
            logger.info("Ingestion finished successfully!")
        else:
            logger.error(f"ERROR: {arg_type} ingestion failure! Please check logs")
            sys.exit(1)
    else:
        logger.error(
            f"ERROR: The workflow request to be ingested does not match specified "
            f"ingestion type {arg_type}. Exiting...."
        )
        sys.exit(1)
