#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
from pathlib import Path

from tqdm import tqdm

from .interfaces import FetchProgressReporter, LocatedFile


class ProgressNonReporter(FetchProgressReporter):
    """
    Reports no progress as files complete. Useful as a default and for testing.
    """

    def __init__(self, total):
        pass

    def complete(self, file: LocatedFile, path: Path):
        """Does nothing with the fact that some files are complete."""


class ProgressBarReporter(FetchProgressReporter):
    """
    Display progress bar as the files are fetched.
    """

    def __init__(self, total):
        self.progress = tqdm(total=total, unit="byte", unit_scale=True, unit_divisor=1024)

    def complete(self, file: LocatedFile, path: Path):
        if path and path.exists():
            self.progress.update(path.stat().st_size)
        else:
            # this must be a "dry run" or something
            self.progress.update(file.size)


class VerboseProgressReporter(FetchProgressReporter):
    """
    Display progress bar as the files are fetched.
    """

    def __init__(self, total):
        self.total = total
        self.sofar = 0

    def complete(self, file: LocatedFile, path: Path):
        if path and path.exists():
            size = path.stat().st_size
            self.sofar += size
            print(f"File {path} written, {size} bytes of {self.total} ({self.sofar / self.total * 100.0}%)")
        else:
            # dry run, so there's no file
            print(f"Pretended to fetch {file.name}")
