#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Make sure our fetcher factory functions fine, friend """

from typing import NamedTuple

from productfetcher.fetcher_factory import ConfiguredFetcherFactory
from productfetcher.fetchers import (
    DryRunFakeFileFetcher,
    NgasDirectCopyFileFetcher,
    NgasModeDetectingFetcher,
    NgasStreamingFileFetcher,
)

# pylint: disable=E0239, E0401, E0402, R0201, R0903


class FakeFile(NamedTuple):
    """Ceci n’est pas une file"""

    size: int
    checksum: int

    # direct copy uses this
    def can_direct_copy(self) -> bool:
        return True


def test_configured_fetcher_factory():
    """
    Various tests of the fetcher factory

    :return:
    """
    fake_file = FakeFile(239823983, -329831113)

    # let's try a few combinations and make sure we find reasonable things
    # first, let's make sure the concurrency value is used properly
    factory = ConfiguredFetcherFactory(concurrency=732)
    assert factory.fetch_plan([]).concurrency == 732

    # let's also make sure that if we specify a lower concurrency value, it gets used instead
    assert factory.fetch_plan([], concurrency=3).concurrency == 3

    # now let's make sure that if we ask for a dry run, that's what we get
    factory = ConfiguredFetcherFactory(dry_run=True)
    assert isinstance(factory.fetch_ngas(fake_file).underlying, DryRunFakeFileFetcher)
    assert isinstance(factory.fetch_oracle_xml(fake_file).underlying, DryRunFakeFileFetcher)

    # now let's make sure that if we ask for streaming or direct copy, we get those
    factory = ConfiguredFetcherFactory(streaming=True)
    assert isinstance(factory.fetch_ngas(fake_file).underlying, NgasStreamingFileFetcher)
    factory = ConfiguredFetcherFactory(direct_copy=True)
    assert isinstance(factory.fetch_ngas(fake_file).underlying, NgasDirectCopyFileFetcher)

    # and let's make sure that if we don't specify streaming or direct copy, we get a detection
    factory = ConfiguredFetcherFactory(streaming=False, direct_copy=False)
    assert isinstance(factory.fetch_ngas(fake_file).underlying, NgasModeDetectingFetcher)
