#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
.. codeauthor:: Nathan Hertz <nhertz@nrao.edu>

File containing definitions for the other half of the capability side of the Workspaces REST API,
concerning capability requests
"""
# pylint: disable=E0401

import json

from capability.views.capability_version import submit_capability_version
from pyramid.httpexceptions import HTTPBadRequest, HTTPNotFound, HTTPPreconditionFailed
from pyramid.request import Request
from pyramid.response import Response
from pyramid.view import view_config

from workspaces.system.enums import RemoteWorkflows


@view_config(route_name="view_capability_request", renderer="json")
def view_capability_request(request: Request) -> Response:
    """
    Pyramid view that accepts a request to view a capability request and responds with the request's info, if it exists
    URL: capability/request/{request_id}

    :param request: GET request
    :return: 200 OK response with JSON-formatted capability request info if found
        or 404 response (HTTPNotFound)
    """
    capability_request = request.capability_info.lookup_capability_request(request.matchdict["request_id"])
    if capability_request:
        return Response(json_body=capability_request.__json__())
    else:
        not_found_msg = (f"Capability request with ID {request.matchdict['request_id']} not found.",)
        return HTTPNotFound(detail=not_found_msg)


@view_config(route_name="view_active_capability_requests", renderer="json")
def view_active_capability_requests(request: Request) -> Response:
    """
    Pyramid view that accepts a request to view all active (i.e. not 'Complete') capability requests
    URL: capability/{capability_name}/request/

    :param request: GET request
    :return: 200 OK response with JSON-formatted capability request info if found
        or 404 response (HTTPNotFound)
    """
    active_capability_requests = request.capability_info.lookup_active_capability_requests(
        capability_name=request.matchdict["capability_name"], filters=request.GET.dict_of_lists()
    )
    if active_capability_requests:
        return Response(json_body=active_capability_requests)
    else:
        not_found_msg = (f"Capability with name {request.matchdict['capability_name']} not found.",)
        return HTTPNotFound(detail=not_found_msg)


@view_config(route_name="view_created_capability_requests", renderer="json")
def view_created_capability_requests(request: Request) -> Response:
    """
    Pyramid view that accepts a request to view all created-but-not-submitted capability requests
    URL: capability/{capability_name}/created-requests/

    :param request: GET request
    :return: 200 OK response with JSON-formatted list of created capability requests if found
        or 404 response (HTTPNotFound) if no capability with given name or no created requests for given capability
    """
    created_capability_requests = request.capability_info.lookup_created_capability_requests(
        request.matchdict["capability_name"]
    )
    if created_capability_requests:
        return Response(json_body=created_capability_requests)
    else:
        not_found_msg = (f"Capability with name {request.matchdict['capability_name']} not found.",)
        return HTTPNotFound(detail=not_found_msg)


@view_config(route_name="create_capability_request", renderer="json")
def create_capability_request(request: Request) -> Response:
    """
    Pyramid view that accepts a request to create a capability request
    URL: capability/{capability_name}/request/create

    :param request: POST request, expecting JSON parameter "parameters"
    :return: 200 OK response with JSON-formatted info of newly created capability request
        or 400 response (HTTPBadRequest) if expected parameters not given
        or 412 response (HTTPPreconditionFailed) if capability with given name does not exist and thus cannot be
        requested
    """
    expected_params = ["parameters"]
    capability_name = request.matchdict["capability_name"]
    params = request.json_body

    if not all([expected in params for expected in expected_params]):
        # JSON params do not contain all expected params
        params_not_given_msg = f"Expected JSON parameters {expected_params}. Received only {params}."
        return HTTPBadRequest(detail=params_not_given_msg)
    elif not request.capability_info.lookup_capability(capability_name):
        # Capability with given name does not exist; can't create a request for it
        does_not_exist_msg = (f"Capability {capability_name} does not exist. Cannot create request.",)
        return HTTPPreconditionFailed(detail=does_not_exist_msg)
    else:
        # TODO: Implement future products
        new_capability_request = request.capability_info.create_capability_request(
            capability_name, parameters=params["parameters"]
        )
        return Response(json_body=new_capability_request.__json__())


@view_config(route_name="create_follow_on_capability_request", renderer="json")
def create_follow_on_capability_request(request: Request) -> Response:
    """
    Pyramid view that accepts a request to create a follow-on capability request from a previous request
    URL: capability/request/{request_id}/followon/{followon_type}

    :param request: POST request
    :return: 200 OK response with JSON-formatted info of newly created capability request
        or 400 response (HTTPBadRequest) if expected parameters not given
        or 412 response (HTTPPreconditionFailed) if capability with given name does not exist and thus cannot be
        requested
    """

    followon_type = request.matchdict["followon_type"]
    request_id = request.matchdict["request_id"]
    capability_info = request.capability_info

    version_number = int(request.matchdict["version"])
    current_version = capability_info.lookup_version(request_id, version_number)
    parameters = current_version.parameters
    user_email = parameters["user_email"]

    previous_workflow_id = current_version.current_execution.current_workflow_request_id
    capability_name = current_version.capability.name

    metadata_content = request.workflow_service.retrieve_file_content(
        followon_type, previous_workflow_id, "metadata.json"
    )
    metadata = json.loads(metadata_content)

    target_dir_name = "/products" if capability_name == "restore_cms" else "/working"
    cms_path = metadata["destinationDirectory"] + target_dir_name

    sdm_id = metadata["fileSetIds"][0] if isinstance(metadata["fileSetIds"], list) else metadata["fileSetIds"]

    new_capability_request = request.capability_info.create_capability_request(
        capability_name=followon_type,
        parameters={
            "user_email": user_email,
            "sdmId": sdm_id,
            "cms_path": cms_path,
        },
    )
    return Response(json_body=new_capability_request.__json__())


@view_config(route_name="edit_capability_request", renderer="json")
def edit_capability_request(request: Request) -> Response:
    """
    Pyramid view that accepts a request to edit a capability request
    URL: capability/{capability_name}/request/{request_id}/edit

    :param request: POST request
    :return: Bad Request response, since it's not yet implemented

    TODO: Implement once CapabilityVersions are supported
    """
    return HTTPBadRequest(detail="Editing capability requests is not yet implemented.")


@view_config(route_name="create_and_submit_capability_request", renderer="json")
def create_and_submit_capability_request(request: Request) -> Response:
    """
    Pyramid view to create and submit a capability request

    :param request: dummy POST request
    :return:
    """
    response = create_capability_request(request)
    capability_request = request.capability_info.lookup_capability_request(response.json_body["id"])
    request.matchdict = {
        "capability_request_id": capability_request.id,
        "version_number": capability_request.current_version.version_number,
    }
    return submit_capability_version(request)


@view_config(route_name="delete_capability_request", renderer="json")
def delete_capability_request(request: Request) -> Response:
    """
    Pyramid view that accepts a request to delete a capability request from the face of the archive
    URL: capability/{capability_name}/request/{request_id}

    :param request: DELETE request
    :return: 200 OK response
        or 412 response (HTTPPreconditionFailed) if capability request with given ID does not exist and thus cannot be
        deleted
        or 400 response (HTTPBadRequest) if capability request with given ID cannot be deleted
    """
    capability_name = request.matchdict["capability_name"]
    request_id = request.matchdict["request_id"]
    capability_request = request.capability_info.lookup_capability_request(request_id)

    if not capability_request:
        # Capability request not found
        does_not_exist_msg = (
            f"Capability request for {capability_name} with ID {request_id} does not exist. Cannot delete."
        )
        return HTTPPreconditionFailed(detail=does_not_exist_msg)

    rows_deleted = request.capability_info.delete_capability_request(request_id)
    if rows_deleted == -1:
        bad_request_msg = (
            f"Capability request for {capability_name} with ID {request_id} cannot be deleted;",
            "it has existing executions.",
        )
        return HTTPBadRequest(detail=bad_request_msg)
    return Response(body=f"Capability request for {capability_name} with ID {request_id} successfully deleted.")


@view_config(route_name="assign_capability_request", renderer="json")
def assign_capability_request(request: Request) -> Response:
    """
    Assign a DA or AOD to a request.

    - URL: capability/request/{request_id}/assign/group/{group}/staff/{user_name}

    :param request: POST request
    :return: 200 OK response
        or 412 response (HTTPPreconditionFailed) if capability request with given ID does not exist and thus cannot be
        updated
        or 400 response (HTTPBadRequest) if capability request assignment fails
    """
    request_id = request.matchdict["request_id"]
    capability_request = request.capability_info.lookup_capability_request(request_id)
    if not capability_request:
        # Capability request not found
        does_not_exist_msg = f"Capability request with ID {request_id} does not exist. Cannot assign."
        return HTTPPreconditionFailed(detail=does_not_exist_msg)

    group = request.matchdict["group"]
    user_name = request.matchdict["user_name"]

    if user_name == "UNASSIGN":
        user_name = None
        staff_member = None
        response_body = f"Unassigned {group} for capability request {request_id}"
    else:
        # Now extract the full QaStaff object from the body:
        expected_params = ["qa_staff"]
        params = request.json_body
        if not all([expected in params for expected in expected_params]):
            # JSON params do not contain all expected params
            params_not_given_msg = f"Expected JSON parameters {expected_params}. Received only {params}."
            return HTTPBadRequest(detail=params_not_given_msg)

        staff_dict = params["qa_staff"]
        staff_member = request.capability_info.get_staff_member(staff_dict)
        response_body = f"{group} {user_name} has been assigned to capability request {request_id}"

    if group == "DA":
        capability_request.assigned_da = user_name
        capability_request.da = staff_member
    elif group == "AOD":
        capability_request.assigned_aod = user_name
        capability_request.aod = staff_member
    else:
        dunno_whatta_do_msg = f"Don't know how to assign request {request_id}"
        return HTTPBadRequest(detail=dunno_whatta_do_msg)

    try:
        request.capability_info.save_entity(capability_request)
    except Exception as exc:
        detail = f"assignment of {user_name} as {group} to request {request_id} failed: {exc}"
        return HTTPBadRequest(detail=detail)

    return Response(json_body={"resp": response_body})
